@@ -334,7 +334,7 @@ def _mirror(axis: np.ndarray) -> np.ndarray:
 
 
 TM3d = TypeVar("TM3d", bound=m3d.Manifold | m3d.CrossSection)
-
+DEBUG_MANIFOLD_3D=False
 
 @dataclass(init=False)
 class RenderContext(Generic[TM3d]):
@@ -418,6 +418,12 @@ class RenderContext(Generic[TM3d]):
     def _apply_and_merge(self, get_type_func: Callable[[], tuple[TM3d, ...]]) -> tuple[TM3d, ...]:
         manifs = self._apply_transforms(get_type_func)
         result_manifs = (sum(manifs[1:], start=manifs[0]),) if len(manifs) > 1 else manifs
+        if DEBUG_MANIFOLD_3D:
+            print(f"Merging {len(result_manifs)} manifolds")
+            for i, m in enumerate(result_manifs):
+                if m.num_vert() == 0:
+                    print(f"Manifold {i} has no vertices")
+                    
         return result_manifs
 
     def _apply_and_merge_helper(
@@ -648,20 +654,20 @@ def set_property(manifold: m3d.Manifold, prop: np.ndarray, prop_index: int) -> m
     # number of properties.
     if prop_index == 0:
         if curr_num_props > len(prop):
-            apply_func = lambda pos, curr_props: np.concatenate((prop, curr_props[prop_len:]))
+            apply_func = lambda pos, curr_props: np.concatenate((prop, curr_props[prop_len:]))  # noqa: E731
         else:
-            apply_func = lambda pos, curr_props: prop
+            apply_func = lambda pos, curr_props: prop  # noqa: E731
     elif num_props > curr_num_props:
-        apply_func = lambda pos, curr_props: np.concatenate((curr_props[:prop_index], prop))
+        apply_func = lambda pos, curr_props: np.concatenate((curr_props[:prop_index], prop))  # noqa: E731
     elif curr_num_props < prop_index:
         padding = np.zeros(prop_index - curr_num_props)
-        apply_func = lambda pos, curr_props: np.concatenate((
+        apply_func = lambda pos, curr_props: np.concatenate((  # noqa: E731
             curr_props[:prop_index],
             padding,
             prop,
         ))
     else:
-        apply_func = lambda pos, curr_props: np.concatenate((
+        apply_func = lambda pos, curr_props: np.concatenate((  # noqa: E731
             curr_props[:prop_index],
             prop,
             curr_props[prop_index + prop_len :],
@@ -986,16 +992,33 @@ class M3dRenderer(RendererBase):
 
         if triangles is None:
             # Convert inputs to numpy arrays if they aren't already
-            verts_array = np.array(verts, dtype=np.float32)
+            verts_array = np.asarray(verts, dtype=np.float32)
 
             # Triangulate each face and collect the indices
             tri_verts = []
+            quad_idxs = []
             for face in faces:
-                tri_verts.extend(triangulate_3d_face(verts_array, [face]))
+                face_verts_n = len(face)
+                if face_verts_n == 3:
+                    tri_verts.extend(face)
+                elif face_verts_n == 4:
+                    quad_idxs.append(face)
+                else:
+                    tri_verts.extend(triangulate_3d_face(verts_array, [face]))
+            if quad_idxs:
+                quad_idxs = np.array(quad_idxs, dtype=np.uint32)
+                choice = choose_quad_triangulation(quad_idxs, verts_array)
+                quad_tris = generate_triangle_indices(quad_idxs, choice)
+                tri_verts = np.concatenate((tri_verts, quad_tris))
+
+            try:
+                return self._mesh(
+                    vert_properties=verts_array, tri_verts=np.asarray(tri_verts, dtype=np.uint32)
+                )
+            except Exception as e:
+                print(f"Error triangulating faces: {e}")
+                raise
 
-            return self._mesh(
-                vert_properties=verts_array, tri_verts=np.array(tri_verts, dtype=np.uint32)
-            )
         elif faces is not None:
             return self._mesh(vert_properties=verts_array, tri_verts=triangles)
         else:
@@ -1457,7 +1480,7 @@ class M3dRenderer(RendererBase):
                 c = np.concatenate((c, [alpha]))
             elif c.shape != (4,):
                 raise ValueError("Color must be a 3 or 4 element numpy array")
-        elif isinstance(c, list):
+        elif isinstance(c, (list, tuple)):
             if len(c) == 3:
                 if alpha is None:
                     alpha = 1.0
@@ -1657,3 +1680,206 @@ def get_polygon_signed_area(poly_verts: np.ndarray) -> float:
     # Use np.roll to get coordinates of the 'next' vertex
     signed_area = 0.5 * np.sum(x * np.roll(y, -1) - np.roll(x, -1) * y)
     return signed_area
+
+import numpy as np
+
+# Helper function (can be defined inside or outside the main function)
+def _cross_norm(v1: np.ndarray, v2: np.ndarray, eps: float = 1e-12) -> np.ndarray:
+    """
+    Calculates normalized cross product robustly, handling zero-length results.
+
+    Args:
+        v1: First batch of vectors (N, 3).
+        v2: Second batch of vectors (N, 3).
+        eps: Tolerance below which norm is considered zero.
+
+    Returns:
+        Normalized cross product vectors (N, 3), or zero vectors where
+        the cross product magnitude is below tolerance.
+    """
+    # Ensure vectors are at least 2D for cross product axis arguments
+    v1 = np.atleast_2d(v1)
+    v2 = np.atleast_2d(v2)
+    cp = np.cross(v1, v2, axisa=1, axisb=1)
+    norm = np.linalg.norm(cp, axis=1, keepdims=True)
+    # Create default zero vector matching the shape of cp
+    default_zero = np.zeros_like(cp)
+    # Perform division, replacing results where norm is near zero with zeros
+    return np.divide(cp, norm, out=default_zero, where=norm > eps)
+
+
+def choose_quad_triangulation(quad_indices: np.ndarray,
+                              vertices: np.ndarray,
+                              epsilon: float = 1e-9) -> np.ndarray:
+    """
+    Determines the preferred triangulation rule (A or B) for multiple quads,
+    using robust checks for non-convex quads and shortest diagonal otherwise.
+
+    Uses a plane-based check to identify internal/external diagonals. If only
+    one diagonal is internal (non-convex case), that diagonal is chosen. If both
+    are internal (convex) or neither is internal (degenerate), the shortest
+    diagonal rule is used.
+
+    Args:
+        quad_indices: A NumPy array of shape (N, 4) and dtype integer.
+                      Each row contains the 4 indices into the 'vertices'
+                      array defining a quad (ordered 0, 1, 2, 3 around perimeter).
+                      N is the number of quads.
+        vertices:     A NumPy array of shape (M, 3) and dtype float.
+                      Contains the 3D coordinates of the M unique vertices.
+        epsilon:      A small tolerance for floating-point comparisons, used
+                      in the internal diagonal and normal checks.
+
+    Returns:
+        A NumPy array of shape (N,) and dtype bool.
+        - True:  Indicates Rule B (diagonal 0-2) should be used.
+                 Chosen if diagonal 0-2 is internal and 1-3 is external, OR
+                 if quad is convex/degenerate and 0-2 is the shorter diagonal.
+        - False: Indicates Rule A (diagonal 1-3) should be used.
+                 Chosen if diagonal 1-3 is internal and 0-2 is external, OR
+                 if quad is convex/degenerate and 1-3 is the shorter or equal diagonal.
+    """
+    # --- Input Validation ---
+    if quad_indices.ndim != 2 or quad_indices.shape[1] != 4:
+        raise ValueError("quad_indices must have shape (N, 4)")
+    if vertices.ndim != 2 or vertices.shape[1] != 3:
+        raise ValueError("vertices must have shape (M, 3)")
+    if not np.issubdtype(quad_indices.dtype, np.integer):
+        raise TypeError("quad_indices must be an integer type array")
+
+    N = quad_indices.shape[0]
+    if N == 0:
+        return np.array([], dtype=bool)
+
+    # --- Get vertex indices and coordinates ---
+    idx0, idx1, idx2, idx3 = quad_indices.T
+    p0 = vertices[idx0] # Shape (N, 3)
+    p1 = vertices[idx1] # Shape (N, 3)
+    p2 = vertices[idx2] # Shape (N, 3)
+    p3 = vertices[idx3] # Shape (N, 3)
+
+    # --- Calculate triangle normals (unnormalized is fine for averaging) ---
+    n012 = np.cross(p1 - p0, p2 - p0, axisa=1, axisb=1)
+    n123 = np.cross(p2 - p1, p3 - p1, axisa=1, axisb=1)
+    n230 = np.cross(p3 - p2, p0 - p2, axisa=1, axisb=1)
+    n301 = np.cross(p0 - p3, p1 - p3, axisa=1, axisb=1)
+
+    # --- Calculate average quad normal vector (robust normalization) ---
+    avg_N_unnorm = n012 + n123 + n230 + n301
+    avg_N_norm = np.linalg.norm(avg_N_unnorm, axis=1, keepdims=True)
+    default_N = np.array([[0., 0., 1.]]) # Default for degenerate quads
+    avg_N = np.divide(avg_N_unnorm, avg_N_norm,
+                      out=np.zeros_like(avg_N_unnorm), where=avg_N_norm > epsilon)
+    zero_norm_mask = (avg_N_norm <= epsilon).flatten()
+    if np.any(zero_norm_mask):
+         avg_N[zero_norm_mask] = np.tile(default_N, (np.sum(zero_norm_mask), 1))
+
+    # --- Check diagonal 0-2 internal validity ---
+    diag02_vec = p2 - p0
+    split_plane_N02 = _cross_norm(diag02_vec, avg_N, eps=epsilon)
+    dist1_plane02 = np.sum((p1 - p0) * split_plane_N02, axis=1)
+    dist3_plane02 = np.sum((p3 - p0) * split_plane_N02, axis=1)
+    # Internal if points on opposite sides (product <= epsilon^2)
+    diag02_internal = (dist1_plane02 * dist3_plane02) <= epsilon**2
+
+    # --- Check diagonal 1-3 internal validity ---
+    diag13_vec = p3 - p1
+    split_plane_N13 = _cross_norm(diag13_vec, avg_N, eps=epsilon)
+    dist0_plane13 = np.sum((p0 - p1) * split_plane_N13, axis=1)
+    dist2_plane13 = np.sum((p2 - p1) * split_plane_N13, axis=1)
+    # Internal if points on opposite sides (product <= epsilon^2)
+    diag13_internal = (dist0_plane13 * dist2_plane13) <= epsilon**2
+
+    # --- Calculate shortest diagonal ---
+    d02_sq = np.sum(diag02_vec**2, axis=1)
+    d13_sq = np.sum(diag13_vec**2, axis=1)
+    # True if B (0-2) is strictly shorter
+    shortest_is_B = d02_sq < d13_sq
+
+    # --- Combine checks into final boolean choice mask ---
+    # Default choice is based on shortest diagonal
+    final_choice_B = shortest_is_B.copy() # True = Use B (0-2)
+
+    # Identify quads where only one diagonal is internal (non-convex cases)
+    only_02_internal = diag02_internal & ~diag13_internal
+    only_13_internal = ~diag02_internal & diag13_internal
+
+    # Force B (True) if only 0-2 is internal
+    final_choice_B[only_02_internal] = True
+    # Force A (False) if only 1-3 is internal
+    final_choice_B[only_13_internal] = False
+    # If both or neither internal, the shortest_is_B choice remains
+
+    return final_choice_B
+
+
+def generate_triangle_indices(quad_indices: np.ndarray, use_rule_b_mask: np.ndarray) -> np.ndarray:
+    """
+    Generates the final triangle indices based on the boolean choice mask.
+
+    Args:
+        quad_indices: A NumPy array of shape (N, 4) and dtype integer.
+        use_rule_b_mask: A NumPy array of shape (N,) and dtype bool.
+                         True means use Rule B (diagonal 0-2).
+                         False means use Rule A (diagonal 1-3).
+
+    Returns:
+        A NumPy array of shape (N * 2, 3) and dtype integer, containing
+        the indices for the two triangles per quad.
+    """
+    N = quad_indices.shape[0]
+    output_dtype = quad_indices.dtype
+    if N == 0:
+        return np.empty((0, 3), dtype=output_dtype) # Return shape (0, 3) for consistency
+
+    # Initialize as an empty 2D array (N*2 triangles, 3 vertices each)
+    output_tris = np.empty((N * 2, 3), dtype=output_dtype)
+
+    # Get original indices for all quads
+    i0 = quad_indices[:, 0]
+    i1 = quad_indices[:, 1]
+    i2 = quad_indices[:, 2]
+    i3 = quad_indices[:, 3]
+
+    # --- Rule A triangles ---
+    rule_a_mask = ~use_rule_b_mask
+    num_a = np.count_nonzero(rule_a_mask)
+
+    if num_a > 0:
+        # Get the indices of the quads where Rule A applies
+        rule_a_row_indices = np.flatnonzero(rule_a_mask)
+        # Calculate the corresponding row indices in the output array
+        rows_a1 = rule_a_row_indices * 2
+        rows_a2 = rows_a1 + 1
+
+        # Assign Rule A triangles directly
+        # Triangle 1 (i0, i1, i3)
+        output_tris[rows_a1, 0] = i0[rule_a_mask]
+        output_tris[rows_a1, 1] = i1[rule_a_mask]
+        output_tris[rows_a1, 2] = i3[rule_a_mask]
+        # Triangle 2 (i1, i2, i3)
+        output_tris[rows_a2, 0] = i1[rule_a_mask]
+        output_tris[rows_a2, 1] = i2[rule_a_mask]
+        output_tris[rows_a2, 2] = i3[rule_a_mask]
+
+    # --- Rule B triangles ---
+    num_b = N - num_a # Faster than counting True mask
+
+    if num_b > 0:
+        # Get the indices of the quads where Rule B applies
+        rule_b_row_indices = np.flatnonzero(use_rule_b_mask)
+        # Calculate the corresponding row indices in the output array
+        rows_b1 = rule_b_row_indices * 2
+        rows_b2 = rows_b1 + 1
+
+        # Assign Rule B triangles directly
+        # Triangle 1 (i0, i1, i2)
+        output_tris[rows_b1, 0] = i0[use_rule_b_mask]
+        output_tris[rows_b1, 1] = i1[use_rule_b_mask]
+        output_tris[rows_b1, 2] = i2[use_rule_b_mask]
+        # Triangle 2 (i0, i2, i3)
+        output_tris[rows_b2, 0] = i0[use_rule_b_mask]
+        output_tris[rows_b2, 1] = i2[use_rule_b_mask]
+        output_tris[rows_b2, 2] = i3[use_rule_b_mask]
+
+    return output_tris
